@@ -24,6 +24,7 @@ import fixtures
 from keystoneauth1 import exceptions as ks_exceptions
 import mock
 from novaclient import exceptions
+from oslo_config import cfg
 from oslo_config import fixture as config_fixture
 from oslotest import base
 from oslotest import log
@@ -41,9 +42,29 @@ class BaseTestCase(base.BaseTestCase):
         super(BaseTestCase, self).setUp()
         self.logger = self.useFixture(log.ConfigureLogging()).logger
         self.conf = self.useFixture(config_fixture.Config())
+        self.conf.config(enable_routed_networks=True)
+        # ctlplane-subnet - config group options
+        self.grp0 = cfg.OptGroup(name='ctlplane-subnet',
+                                 title='ctlplane-subnet')
+        self.opts = [cfg.StrOpt('cidr'),
+                     cfg.StrOpt('dhcp_start'),
+                     cfg.StrOpt('dhcp_end'),
+                     cfg.StrOpt('inspection_iprange'),
+                     cfg.StrOpt('gateway'),
+                     cfg.BoolOpt('masquerade')]
+        self.conf.register_opts(self.opts, group=self.grp0)
+        self.grp1 = cfg.OptGroup(name='subnet1', title='subnet1')
+        self.gtp2 = cfg.OptGroup(name='subnet2', title='subnet2')
+        self.conf.config(cidr='192.168.24.0/24',
+                         dhcp_start='192.168.24.5', dhcp_end='192.168.24.24',
+                         inspection_iprange='192.168.24.100,192.168.24.120',
+                         gateway='192.168.24.1', masquerade=True,
+                         group='ctlplane-subnet')
 
 
 class TestUndercloud(BaseTestCase):
+    @mock.patch(
+        'instack_undercloud.undercloud._load_subnets_config_groups')
     @mock.patch('instack_undercloud.undercloud._handle_upgrade_fact')
     @mock.patch('instack_undercloud.undercloud._configure_logging')
     @mock.patch('instack_undercloud.undercloud._validate_configuration')
@@ -63,7 +84,7 @@ class TestUndercloud(BaseTestCase):
                      mock_run_clean_all, mock_run_yum_update, mock_run_orc,
                      mock_post_config, mock_run_command,
                      mock_validate_configuration, mock_configure_logging,
-                     mock_upgrade_fact):
+                     mock_upgrade_fact, mock_load_subnets_config_groups):
         fake_env = mock.MagicMock()
         mock_generate_environment.return_value = fake_env
         undercloud.install('.')
@@ -77,6 +98,8 @@ class TestUndercloud(BaseTestCase):
         mock_die_tuskar_die.assert_not_called()
         mock_run_validation_groups.assert_not_called()
 
+    @mock.patch(
+        'instack_undercloud.undercloud._load_subnets_config_groups')
     @mock.patch('instack_undercloud.undercloud._handle_upgrade_fact')
     @mock.patch('instack_undercloud.undercloud._configure_logging')
     @mock.patch('instack_undercloud.undercloud._validate_configuration')
@@ -96,7 +119,8 @@ class TestUndercloud(BaseTestCase):
                              mock_run_yum_clean_all, mock_run_yum_update,
                              mock_run_orc, mock_post_config, mock_run_command,
                              mock_validate_configuration,
-                             mock_configure_logging, mock_upgrade_fact):
+                             mock_configure_logging, mock_upgrade_fact,
+                             mock_load_subnets_config_groups):
         fake_env = mock.MagicMock()
         mock_generate_environment.return_value = fake_env
         undercloud.install('.', upgrade=True)
@@ -110,6 +134,8 @@ class TestUndercloud(BaseTestCase):
         mock_die_tuskar_die.assert_called_once()
         mock_run_validation_groups.assert_called_once()
 
+    @mock.patch(
+        'instack_undercloud.undercloud._load_subnets_config_groups')
     @mock.patch('instack_undercloud.undercloud._handle_upgrade_fact')
     @mock.patch('instack_undercloud.undercloud._configure_logging')
     @mock.patch('instack_undercloud.undercloud._validate_configuration')
@@ -132,7 +158,8 @@ class TestUndercloud(BaseTestCase):
                                        mock_post_config, mock_run_command,
                                        mock_validate_configuration,
                                        mock_configure_logging,
-                                       mock_upgrade_fact):
+                                       mock_upgrade_fact,
+                                       mock_load_subnets_config_groups):
         self.conf.config(hieradata_override='override.yaml')
         with open(os.path.expanduser('~/override.yaml'), 'w') as f:
             f.write('Something\n')
@@ -634,6 +661,130 @@ class TestGenerateEnvironment(BaseTestCase):
         env = undercloud._generate_environment('.')
         self.assertNotIn(env, 'DIB_YUM_REPO_CONF')
 
+    def test_inspection_ip_single_subnet(self):
+        env = undercloud._generate_environment('.')
+        reference = [{"tag": "ctlplane-subnet", "gateway": "192.168.24.1",
+                      "ip_range": "192.168.24.100,192.168.24.120",
+                      "netmask": "255.255.255.0"}]
+        actual = json.loads(env['INSPECTION_SUBNETS'])
+        self.assertEqual(reference, actual)
+
+    def test_inspection_ip_multiple_subnets(self):
+        self.conf.config(subnets=['subnet1', 'subnet2'])
+        self.conf.config(local_subnet='subnet1')
+        self.conf.register_opts(self.opts, group=self.grp1)
+        self.conf.register_opts(self.opts, group=self.gtp2)
+        self.conf.config(cidr='192.168.10.0/24', dhcp_start='192.168.10.10',
+                         dhcp_end='192.168.10.99',
+                         inspection_iprange='192.168.10.100,192.168.10.189',
+                         gateway='192.168.10.254', masquerade=True,
+                         group='subnet1')
+        self.conf.config(cidr='192.168.20.0/24', dhcp_start='192.168.20.10',
+                         dhcp_end='192.168.20.99',
+                         inspection_iprange='192.168.20.100,192.168.20.189',
+                         gateway='192.168.20.254', masquerade=True,
+                         group='subnet2')
+        env = undercloud._generate_environment('.')
+        reference = [{"tag": "subnet1", "gateway": "192.168.10.254",
+                      "ip_range": "192.168.10.100,192.168.10.189",
+                      "netmask": "255.255.255.0"},
+                     {"tag": "subnet2", "gateway": "192.168.20.254",
+                      "ip_range": "192.168.20.100,192.168.20.189",
+                      "netmask": "255.255.255.0"}]
+        actual = json.loads(env['INSPECTION_SUBNETS'])
+        self.assertEqual(reference, actual)
+
+    def test_subnets_static_routes(self):
+        self.conf.config(subnets=['ctlplane-subnet', 'subnet1', 'subnet2'])
+        self.conf.register_opts(self.opts, group=self.grp1)
+        self.conf.register_opts(self.opts, group=self.gtp2)
+        self.conf.config(cidr='192.168.24.0/24',
+                         dhcp_start='192.168.24.5', dhcp_end='192.168.24.24',
+                         inspection_iprange='192.168.24.100,192.168.24.120',
+                         gateway='192.168.24.1', masquerade=True,
+                         group='ctlplane-subnet')
+        self.conf.config(cidr='192.168.10.0/24', dhcp_start='192.168.10.10',
+                         dhcp_end='192.168.10.99',
+                         inspection_iprange='192.168.10.100,192.168.10.189',
+                         gateway='192.168.10.254', masquerade=True,
+                         group='subnet1')
+        self.conf.config(cidr='192.168.20.0/24', dhcp_start='192.168.20.10',
+                         dhcp_end='192.168.20.99',
+                         inspection_iprange='192.168.20.100,192.168.20.189',
+                         gateway='192.168.20.254', masquerade=True,
+                         group='subnet2')
+        env = undercloud._generate_environment('.')
+        reference = [{"ip_netmask": "192.168.10.0/24",
+                      "next_hop": "192.168.24.1"},
+                     {"ip_netmask": "192.168.20.0/24",
+                      "next_hop": "192.168.24.1"}]
+        actual = json.loads(env['SUBNETS_STATIC_ROUTES'])
+        self.assertEqual(reference, actual)
+
+    def test_subnets_subnets_cidr_nat_rules(self):
+        self.conf.config(subnets=['ctlplane-subnet', 'subnet1', 'subnet2'])
+        self.conf.register_opts(self.opts, group=self.grp1)
+        self.conf.register_opts(self.opts, group=self.gtp2)
+        self.conf.config(cidr='192.168.24.0/24',
+                         dhcp_start='192.168.24.5', dhcp_end='192.168.24.24',
+                         inspection_iprange='192.168.24.100,192.168.24.120',
+                         gateway='192.168.24.1', group='ctlplane-subnet')
+        self.conf.config(cidr='192.168.10.0/24', dhcp_start='192.168.10.10',
+                         dhcp_end='192.168.10.99',
+                         inspection_iprange='192.168.10.100,192.168.10.189',
+                         gateway='192.168.10.254', group='subnet1')
+        self.conf.config(cidr='192.168.20.0/24', dhcp_start='192.168.20.10',
+                         dhcp_end='192.168.20.99',
+                         inspection_iprange='192.168.20.100,192.168.20.189',
+                         gateway='192.168.20.254', group='subnet2')
+
+        env = undercloud._generate_environment('.')
+        reference = ('"140 destination ctlplane-subnet cidr nat": '
+                     '{"chain": "FORWARD", "destination": "192.168.24.0/24", '
+                     '"proto": "all", "action": "accept"}'
+                     '\n  "140 source ctlplane-subnet cidr nat": '
+                     '{"chain": "FORWARD", "source": "192.168.24.0/24", '
+                     '"proto": "all", "action": "accept"}'
+                     '\n  "140 destination subnet1 cidr nat": '
+                     '{"chain": "FORWARD", "destination": "192.168.10.0/24", '
+                     '"proto": "all", "action": "accept"}'
+                     '\n  "140 source subnet1 cidr nat": '
+                     '{"chain": "FORWARD", "source": "192.168.10.0/24", '
+                     '"proto": "all", "action": "accept"}'
+                     '\n  "140 destination subnet2 cidr nat": '
+                     '{"chain": "FORWARD", "destination": "192.168.20.0/24", '
+                     '"proto": "all", "action": "accept"}'
+                     '\n  "140 source subnet2 cidr nat": '
+                     '{"chain": "FORWARD", "source": "192.168.20.0/24", '
+                     '"proto": "all", "action": "accept"}')
+        actual = env['SUBNETS_CIDR_NAT_RULES']
+        self.assertEqual(reference, actual)
+
+    def test_masquerade_networks(self):
+        self.conf.config(subnets=['ctlplane-subnet', 'subnet1', 'subnet2'])
+        self.conf.register_opts(self.opts, group=self.grp1)
+        self.conf.register_opts(self.opts, group=self.gtp2)
+        self.conf.config(cidr='192.168.24.0/24',
+                         dhcp_start='192.168.24.5', dhcp_end='192.168.24.24',
+                         inspection_iprange='192.168.24.100,192.168.24.120',
+                         gateway='192.168.24.1', masquerade=True,
+                         group='ctlplane-subnet')
+        self.conf.config(cidr='192.168.10.0/24', dhcp_start='192.168.10.10',
+                         dhcp_end='192.168.10.99',
+                         inspection_iprange='192.168.10.100,192.168.10.189',
+                         gateway='192.168.10.254', masquerade=True,
+                         group='subnet1')
+        self.conf.config(cidr='192.168.20.0/24', dhcp_start='192.168.20.10',
+                         dhcp_end='192.168.20.99',
+                         inspection_iprange='192.168.20.100,192.168.20.189',
+                         gateway='192.168.20.254', masquerade=True,
+                         group='subnet2')
+
+        env = undercloud._generate_environment('.')
+        reference = ['192.168.24.0/24', '192.168.10.0/24', '192.168.20.0/24']
+        actual = json.loads(env['MASQUERADE_NETWORKS'])
+        self.assertEqual(reference, actual)
+
 
 class TestWritePasswordFile(BaseTestCase):
     def test_normal(self):
@@ -769,10 +920,13 @@ class TestConfigureSshKeys(base.BaseTestCase):
         self._test_configure_ssh_keys(mock_eui, False)
 
 
-class TestPostConfig(base.BaseTestCase):
+class TestPostConfig(BaseTestCase):
     @mock.patch('os_client_config.make_client')
     @mock.patch('instack_undercloud.undercloud._migrate_to_convergence')
     @mock.patch('instack_undercloud.undercloud._ensure_node_resource_classes')
+    @mock.patch(
+      'instack_undercloud.undercloud._config_neutron_segments_and_subnets')
+    @mock.patch('instack_undercloud.undercloud._ensure_neutron_network')
     @mock.patch('instack_undercloud.undercloud._member_role_exists')
     @mock.patch('instack_undercloud.undercloud._get_session')
     @mock.patch('ironicclient.client.get_client', autospec=True)
@@ -790,6 +944,8 @@ class TestPostConfig(base.BaseTestCase):
                          mock_copy_stackrc, mock_delete, mock_mistral_client,
                          mock_swift_client, mock_nova_client, mock_ir_client,
                          mock_get_session, mock_member_role_exists,
+                         mock_ensure_neutron_network,
+                         mock_config_neutron_segments_and_subnets,
                          mock_resource_classes, mock_migrate_to_convergence,
                          mock_make_client):
         instack_env = {
@@ -1284,6 +1440,163 @@ class TestPostConfig(base.BaseTestCase):
         mock_cmce.assert_called_once_with(instack_env, mock_mistral)
         mock_create.assert_called_once_with(mock_mistral, ['hut8'])
 
+    def _neutron_mocks(self):
+        mock_sdk = mock.MagicMock()
+        mock_sdk.network.create_network = mock.Mock()
+        mock_sdk.network.create_segment = mock.Mock()
+        mock_sdk.network.update_segment = mock.Mock()
+        mock_sdk.network.delete_segment = mock.Mock()
+        mock_sdk.network.create_subnet = mock.Mock()
+        mock_sdk.network.update_subnet = mock.Mock()
+        return mock_sdk
+
+    def test_network_create(self):
+        mock_sdk = self._neutron_mocks()
+        mock_sdk.network.networks.return_value = iter([])
+        segment_mock = mock.Mock()
+        mock_sdk.network.segments.return_value = iter([segment_mock])
+        undercloud._ensure_neutron_network(mock_sdk)
+        mock_sdk.network.create_network.assert_called_with(
+          name='ctlplane', provider_network_type='flat',
+          provider_physical_network='ctlplane')
+
+    def test_delete_default_segment(self):
+        mock_sdk = self._neutron_mocks()
+        mock_sdk.network.networks.return_value = iter([])
+        segment_mock = mock.Mock()
+        mock_sdk.network.segments.return_value = iter([segment_mock])
+        undercloud._ensure_neutron_network(mock_sdk)
+        mock_sdk.network.delete_segment.assert_called_with(
+          segment_mock.id)
+
+    def test_network_exists(self):
+        mock_sdk = self._neutron_mocks()
+        mock_sdk.network.networks.return_value = iter(['ctlplane'])
+        undercloud._ensure_neutron_network(mock_sdk)
+        mock_sdk.network.create_network.assert_not_called()
+
+    def test_segment_create(self):
+        mock_sdk = self._neutron_mocks()
+        undercloud._neutron_segment_create(mock_sdk, 'ctlplane-subnet',
+                                           'network_id', 'ctlplane')
+        mock_sdk.network.create_segment.assert_called_with(
+          name='ctlplane-subnet', network_id='network_id',
+          physical_network='ctlplane', network_type='flat')
+
+    def test_segment_update(self):
+        mock_sdk = self._neutron_mocks()
+        undercloud._neutron_segment_update(mock_sdk,
+                                           'network_id', 'ctlplane-subnet')
+        mock_sdk.network.update_segment.assert_called_with(
+            'network_id', name='ctlplane-subnet')
+
+    def test_subnet_create(self):
+        mock_sdk = self._neutron_mocks()
+        host_routes = [{'destination': '169.254.169.254/32',
+                        'nexthop': '192.168.24.1'}]
+        allocation_pool = [{'start': '192.168.24.5', 'end': '192.168.24.24'}]
+        undercloud._neutron_subnet_create(mock_sdk, 'network_id',
+                                          '192.168.24.0/24', '192.168.24.1',
+                                          host_routes, allocation_pool,
+                                          'ctlplane-subnet', 'segment_id')
+        mock_sdk.network.create_subnet.assert_called_with(
+          name='ctlplane-subnet', cidr='192.168.24.0/24',
+          gateway_ip='192.168.24.1', host_routes=host_routes, enable_dhcp=True,
+          ip_version='4', allocation_pools=allocation_pool,
+          network_id='network_id', segment_id='segment_id')
+
+    def test_subnet_update(self):
+        mock_sdk = self._neutron_mocks()
+        host_routes = [{'destination': '169.254.169.254/32',
+                        'nexthop': '192.168.24.1'}]
+        allocation_pool = [{'start': '192.168.24.5', 'end': '192.168.24.24'}]
+        undercloud._neutron_subnet_update(mock_sdk, 'subnet_id',
+                                          '192.168.24.1', host_routes,
+                                          allocation_pool, 'ctlplane-subnet')
+        mock_sdk.network.update_subnet.assert_called_with(
+          'subnet_id', name='ctlplane-subnet', gateway_ip='192.168.24.1',
+          host_routes=host_routes, allocation_pools=allocation_pool)
+
+    @mock.patch('instack_undercloud.undercloud._neutron_subnet_update')
+    @mock.patch('instack_undercloud.undercloud._get_subnet')
+    def test_no_neutron_segments_if_pre_segments_undercloud(
+      self, mock_get_subnet, mock_neutron_subnet_update):
+        mock_sdk = self._neutron_mocks()
+        mock_subnet = mock.Mock()
+        mock_subnet.segment_id = None
+        mock_get_subnet.return_value = mock_subnet
+        undercloud._config_neutron_segments_and_subnets(mock_sdk,
+                                                        'ctlplane_id')
+        mock_sdk.network.create_segment.assert_not_called()
+        mock_sdk.network.update_segment.assert_not_called()
+        mock_neutron_subnet_update.called_once()
+
+    @mock.patch('instack_undercloud.undercloud._neutron_segment_create')
+    @mock.patch('instack_undercloud.undercloud._neutron_subnet_create')
+    @mock.patch('instack_undercloud.undercloud._get_segment')
+    @mock.patch('instack_undercloud.undercloud._get_subnet')
+    def test_segment_and_subnet_create(self, mock_get_subnet, mock_get_segment,
+                                       mock_neutron_subnet_create,
+                                       mock_neutron_segment_create):
+        mock_sdk = self._neutron_mocks()
+        mock_get_subnet.return_value = None
+        mock_get_segment.return_value = None
+        undercloud._config_neutron_segments_and_subnets(mock_sdk,
+                                                        'ctlplane_id')
+        mock_neutron_segment_create.assert_called_with(
+          mock_sdk, 'ctlplane-subnet', 'ctlplane_id', 'ctlplane')
+        host_routes = [{'destination': '169.254.169.254/32',
+                        'nexthop': '192.168.24.1'}]
+        allocation_pool = [{'start': '192.168.24.5', 'end': '192.168.24.24'}]
+        mock_neutron_subnet_create.assert_called_with(
+          mock_sdk, 'ctlplane_id', '192.168.24.0/24', '192.168.24.1',
+          host_routes, allocation_pool, 'ctlplane-subnet',
+          mock_neutron_segment_create().id)
+
+    @mock.patch('instack_undercloud.undercloud._neutron_segment_update')
+    @mock.patch('instack_undercloud.undercloud._neutron_subnet_update')
+    @mock.patch('instack_undercloud.undercloud._get_segment')
+    @mock.patch('instack_undercloud.undercloud._get_subnet')
+    def test_segment_and_subnet_update(self, mock_get_subnet, mock_get_segment,
+                                       mock_neutron_subnet_update,
+                                       mock_neutron_segment_update):
+        mock_sdk = self._neutron_mocks()
+        mock_subnet = mock.Mock()
+        mock_subnet.id = 'subnet_id'
+        mock_subnet.segment_id = 'segment_id'
+        mock_get_subnet.return_value = mock_subnet
+        mock_segment = mock.Mock()
+        mock_get_segment.return_value = mock_segment
+        mock_segment.id = 'segment_id'
+        undercloud._config_neutron_segments_and_subnets(mock_sdk,
+                                                        'ctlplane_id')
+        mock_neutron_segment_update.assert_called_with(
+          mock_sdk, mock_subnet.segment_id, 'ctlplane-subnet')
+        host_routes = [{'destination': '169.254.169.254/32',
+                        'nexthop': '192.168.24.1'}]
+        allocation_pool = [{'start': '192.168.24.5', 'end': '192.168.24.24'}]
+        mock_neutron_subnet_update.assert_called_with(
+          mock_sdk, 'subnet_id', '192.168.24.1', host_routes,
+          allocation_pool, 'ctlplane-subnet')
+
+    @mock.patch('instack_undercloud.undercloud._get_segment')
+    @mock.patch('instack_undercloud.undercloud._get_subnet')
+    def test_local_subnet_cidr_conflict(self, mock_get_subnet,
+                                        mock_get_segment):
+        mock_sdk = self._neutron_mocks()
+        mock_sdk = self._neutron_mocks()
+        mock_subnet = mock.Mock()
+        mock_subnet.id = 'subnet_id'
+        mock_subnet.segment_id = 'existing_segment_id'
+        mock_get_subnet.return_value = mock_subnet
+        mock_segment = mock.Mock()
+        mock_get_segment.return_value = mock_segment
+        mock_segment.id = 'segment_id'
+        self.assertRaises(
+          RuntimeError,
+          undercloud._config_neutron_segments_and_subnets, [mock_sdk],
+          ['ctlplane_id'])
+
 
 class TestUpgradeFact(base.BaseTestCase):
     @mock.patch('instack_undercloud.undercloud._run_command')
