@@ -21,6 +21,7 @@ import glob
 import hashlib
 import json
 import logging
+import netaddr
 import os
 import platform
 import re
@@ -37,6 +38,7 @@ from keystoneauth1 import exceptions as ks_exceptions
 from keystoneclient import discover
 import keystoneauth1.identity.generic as ks_auth
 from mistralclient.api import client as mistralclient
+from mistralclient.api import base as mistralclient_exc
 from novaclient import client as novaclient
 from novaclient import exceptions
 import os_client_config
@@ -111,7 +113,20 @@ log can be found at %(log_file)s.
 # We need 8 GB, leave a little room for variation in what 8 GB means on
 # different platforms.
 REQUIRED_MB = 7680
+# Control plane network name
+PHYSICAL_NETWORK = 'ctlplane'
+SUBNETS_DEFAULT = ['ctlplane-subnet']
 
+# Deprecated options
+_deprecated_opt_network_gateway = [cfg.DeprecatedOpt(
+  'network_gateway', group='DEFAULT')]
+_deprecated_opt_network_cidr = [cfg.DeprecatedOpt(
+  'network_cidr', group='DEFAULT')]
+_deprecated_opt_dhcp_start = [cfg.DeprecatedOpt(
+  'dhcp_start', group='DEFAULT')]
+_deprecated_opt_dhcp_end = [cfg.DeprecatedOpt('dhcp_end', group='DEFAULT')]
+_deprecated_opt_inspection_iprange = [cfg.DeprecatedOpt(
+  'inspection_iprange', group='DEFAULT')]
 
 # When adding new options to the lists below, make sure to regenerate the
 # sample config by running "tox -e genconfig" in the project root.
@@ -133,12 +148,6 @@ _opts = [
                      'local_interface, with the netmask defined by the '
                      'prefix portion of the value.')
                ),
-    cfg.StrOpt('network_gateway',
-               default='192.168.24.1',
-               help=('Network gateway for the Neutron-managed network for '
-                     'Overcloud instances. This should match the local_ip '
-                     'above when using masquerading.')
-               ),
     cfg.StrOpt('undercloud_public_host',
                deprecated_name='undercloud_public_vip',
                default='192.168.24.2',
@@ -164,6 +173,36 @@ _opts = [
                      'The overcloud parameter "CloudDomain" must be set to a '
                      'matching value.')
                ),
+    cfg.ListOpt('subnets',
+                default=SUBNETS_DEFAULT,
+                help=('List of routed network subnets for provisioning '
+                      'and introspection. Comma separated list of names/tags. '
+                      'For each network a section/group needs to be added to '
+                      'the configuration file with these parameters set: '
+                      'cidr, dhcp_start, dhcp_end, inspection_iprange, '
+                      'gateway and masquerade_network.'
+                      '\n\n'
+                      'Example:\n\n'
+                      'subnets = subnet1,subnet2\n'
+                      '\n'
+                      'An example section/group in config file:\n'
+                      '\n'
+                      '[subnet1]\n'
+                      'cidr = 192.168.10.0/24\n'
+                      'dhcp_start = 192.168.10.100\n'
+                      'dhcp_end = 192.168.10.200\n'
+                      'inspection_iprange = 192.168.10.20,192.168.10.90\n'
+                      'gateway = 192.168.10.254\n'
+                      'masquerade_network = True'
+                      '\n'
+                      '[subnet2]\n'
+                      '. . .\n')),
+    cfg.StrOpt('local_subnet',
+               default=SUBNETS_DEFAULT[0],
+               help=('Name of the local subnet, where the PXE boot and DHCP '
+                     'interfaces for overcloud instances is located. The IP '
+                     'address of the local_ip/local_interface should reside '
+                     'in this subnet.')),
     cfg.StrOpt('undercloud_service_certificate',
                default='',
                help=('Certificate file to use for OpenStack service SSL '
@@ -208,28 +247,17 @@ _opts = [
                default=1500,
                help=('MTU to use for the local_interface.')
                ),
-    cfg.StrOpt('network_cidr',
-               default='192.168.24.0/24',
-               help=('Network CIDR for the Neutron-managed network for '
-                     'Overcloud instances. This should be the subnet used '
-                     'for PXE booting.')
-               ),
     cfg.StrOpt('masquerade_network',
                default='192.168.24.0/24',
+               deprecated_for_removal=True,
+               deprecated_reason=('With support for routed networks, '
+                                  'masquerading of the provisioning networks '
+                                  'is moved to a boolean option for each '
+                                  'subnet.'),
                help=('Network that will be masqueraded for external access, '
                      'if required. This should be the subnet used for PXE '
                      'booting.')
                ),
-    cfg.StrOpt('dhcp_start',
-               default='192.168.24.5',
-               help=('Start of DHCP allocation range for PXE and DHCP of '
-                     'Overcloud instances.')
-               ),
-    cfg.StrOpt('dhcp_end',
-               default='192.168.24.24',
-               help=('End of DHCP allocation range for PXE and DHCP of '
-                     'Overcloud instances.')
-               ),
     cfg.StrOpt('hieradata_override',
                default='',
                help=('Path to hieradata override file. If set, the file will '
@@ -254,14 +282,6 @@ _opts = [
                help=('Network interface on which inspection dnsmasq will '
                      'listen.  If in doubt, use the default value.')
                ),
-    cfg.StrOpt('inspection_iprange',
-               default='192.168.24.100,192.168.24.120',
-               deprecated_name='discovery_iprange',
-               help=('Temporary IP range that will be given to nodes during '
-                     'the inspection process.  Should not overlap with the '
-                     'range defined by dhcp_start and dhcp_end, but should '
-                     'be in the same network.')
-               ),
     cfg.BoolOpt('inspection_extras',
                 default=True,
                 help=('Whether to enable extra hardware collection during '
@@ -375,6 +395,48 @@ _opts = [
                help=('An optional docker \'registry-mirror\' that will be'
                      'configured in /etc/docker/daemon.json.')
                ),
+    cfg.ListOpt('additional_architectures',
+                default=[],
+                help=('List of additional architectures enabled in your cloud '
+                      'environment. The list of supported values is: %s'
+                      % ' '.join(validator.SUPPORTED_ARCHITECTURES))),
+    cfg.BoolOpt('enable_routed_networks',
+                default=False,
+                help=('Enable support for routed ctlplane networks.')),
+]
+
+# Routed subnets
+_subnets_opts = [
+    cfg.StrOpt('cidr',
+               default='192.168.24.0/24',
+               deprecated_opts=_deprecated_opt_network_cidr,
+               help=('Network CIDR for the Neutron-managed subnet for '
+                     'Overcloud instances.')),
+    cfg.StrOpt('dhcp_start',
+               default='192.168.24.5',
+               deprecated_opts=_deprecated_opt_dhcp_start,
+               help=('Start of DHCP allocation range for PXE and DHCP of '
+                     'Overcloud instances on this network.')),
+    cfg.StrOpt('dhcp_end',
+               default='192.168.24.24',
+               deprecated_opts=_deprecated_opt_dhcp_end,
+               help=('End of DHCP allocation range for PXE and DHCP of '
+                     'Overcloud instances on this network.')),
+    cfg.StrOpt('inspection_iprange',
+               default='192.168.24.100,192.168.24.120',
+               deprecated_opts=_deprecated_opt_inspection_iprange,
+               help=('Temporary IP range that will be given to nodes on this '
+                     'network during the inspection process. Should not '
+                     'overlap with the range defined by dhcp_start and '
+                     'dhcp_end, but should be in the same ip subnet.')),
+    cfg.StrOpt('gateway',
+               default='192.168.24.1',
+               deprecated_opts=_deprecated_opt_network_gateway,
+               help=('Network gateway for the Neutron-managed network for '
+                     'Overcloud instances on this network.')),
+    cfg.BoolOpt('masquerade',
+                default=False,
+                help=('The network will be masqueraded for external access.')),
 ]
 
 # Passwords, tokens, hashes
@@ -504,8 +566,15 @@ CONF.register_opts(_opts)
 CONF.register_opts(_auth_opts, group='auth')
 
 
+def _load_subnets_config_groups():
+    for group in CONF.subnets:
+        g = cfg.OptGroup(name=group, title=group)
+        CONF.register_opts(_subnets_opts, group=g)
+
+
 def list_opts():
     return [(None, copy.deepcopy(_opts)),
+            (SUBNETS_DEFAULT[0], copy.deepcopy(_subnets_opts)),
             ('auth', copy.deepcopy(_auth_opts)),
             ]
 
@@ -708,14 +777,44 @@ def _check_sysctl():
         raise RuntimeError('Missing sysctl options')
 
 
+def _cidr_overlaps(a, b):
+    return a.first <= b.last and b.first <= a.last
+
+
 def _validate_network():
     def error_handler(message):
         LOG.error('Undercloud configuration validation failed: %s', message)
         raise validator.FailedValidation(message)
 
+    if (len(CONF.subnets) > 1 and not CONF.enable_routed_networks):
+        message = ('Multiple subnets specified: %s but routed networks are '
+                   'not enabled.' % CONF.subnets)
+        error_handler(message)
+
     params = {opt.name: CONF[opt.name] for opt in _opts}
+    # Get parameters of "local_subnet", pass to validator to ensure parameters
+    # such as "local_ip", "undercloud_public_host" and "undercloud_admin_host"
+    # are valid
+    local_subnet_opts = CONF.get(CONF.local_subnet)
+    params.update({opt.name: local_subnet_opts[opt.name]
+                   for opt in _subnets_opts})
     validator.validate_config(params, error_handler)
 
+    # Validate subnet parameters
+    subnet_cidrs = []
+    for subnet in CONF.subnets:
+        subnet_opts = CONF.get(subnet)
+        params = {opt.name: subnet_opts[opt.name] for opt in _subnets_opts}
+
+        if any(_cidr_overlaps(x, netaddr.IPNetwork(subnet_opts.cidr))
+               for x in subnet_cidrs):
+            message = ('CIDR of %s, %s, overlaps with another subnet.' %
+                       (subnet, subnet_opts.cidr))
+            error_handler(message)
+        subnet_cidrs.append(netaddr.IPNetwork(subnet_opts.cidr))
+
+        validator.validate_subnet(subnet, params, error_handler)
+
 
 def _validate_no_ip_change():
     """Disallow provisioning interface IP changes
@@ -741,7 +840,7 @@ def _validate_no_ip_change():
     if existing_ip != CONF.local_ip:
         message = ('Changing the local_ip is not allowed.  Existing IP: '
                    '%s, Configured IP: %s') % (existing_ip,
-                                               CONF.network_cidr)
+                                               CONF.local_ip)
         LOG.error(message)
         raise validator.FailedValidation(message)
 
@@ -762,6 +861,16 @@ def _validate_passwords_file():
         raise validator.FailedValidation(message)
 
 
+def _validate_architecure_options():
+    def error_handler(message):
+        LOG.error('Undercloud configuration validation failed: %s', message)
+        raise validator.FailedValidation(message)
+
+    params = {opt.name: CONF[opt.name] for opt in _opts}
+    validator._validate_additional_architectures(params, error_handler)
+    validator._validate_ppc64le_exclusive_opts(params, error_handler)
+
+
 def _validate_configuration():
     try:
         _check_hostname()
@@ -770,6 +879,7 @@ def _validate_configuration():
         _validate_network()
         _validate_no_ip_change()
         _validate_passwords_file()
+        _validate_architecure_options()
     except RuntimeError as e:
         LOG.error('An error occurred during configuration validation, '
                   'please check your host configuration and try again. '
@@ -1064,7 +1174,9 @@ class InstackEnvironment(dict):
                     'ENABLED_BOOT_INTERFACES', 'ENABLED_POWER_INTERFACES',
                     'ENABLED_RAID_INTERFACES', 'ENABLED_VENDOR_INTERFACES',
                     'ENABLED_MANAGEMENT_INTERFACES', 'SYSCTL_SETTINGS',
-                    'LOCAL_IP_WRAPPED'}
+                    'LOCAL_IP_WRAPPED', 'ENABLE_ARCHITECTURE_PPC64LE',
+                    'INSPECTION_SUBNETS', 'SUBNETS_CIDR_NAT_RULES',
+                    'SUBNETS_STATIC_ROUTES', 'MASQUERADE_NETWORKS'}
     """The variables we calculate in _generate_environment call."""
 
     PUPPET_KEYS = DYNAMIC_KEYS | {opt.name.upper() for _, group in list_opts()
@@ -1167,6 +1279,60 @@ def _process_drivers_and_hardware_types(instack_env):
     instack_env['ENABLED_POWER_INTERFACES'] = _make_list(mgmt_interfaces)
 
 
+def _generate_masquerade_networks():
+    env_list = []
+    for subnet in CONF.subnets:
+        s = CONF.get(subnet)
+        if s.masquerade:
+            env_list.append(s.cidr)
+
+    # NOTE(hjensas): Remove once deprecated masquerade_network option is gone
+    if CONF.masquerade_network and (CONF.masquerade_network not in env_list):
+        env_list.append(CONF.masquerade_network)
+
+    return json.dumps(env_list)
+
+
+def _generate_inspection_subnets():
+    env_list = []
+    for subnet in CONF.subnets:
+        env_dict = {}
+        s = CONF.get(subnet)
+        env_dict['tag'] = subnet
+        env_dict['ip_range'] = s.inspection_iprange
+        env_dict['netmask'] = str(netaddr.IPNetwork(s.cidr).netmask)
+        env_dict['gateway'] = s.gateway
+        env_list.append(env_dict)
+    return json.dumps(env_list)
+
+
+def _generate_subnets_static_routes():
+    env_list = []
+    local_router = CONF.get(CONF.local_subnet).gateway
+    for subnet in CONF.subnets:
+        if subnet == str(CONF.local_subnet):
+            continue
+        s = CONF.get(subnet)
+        env_list.append({'ip_netmask': s.cidr,
+                         'next_hop': local_router})
+    return json.dumps(env_list)
+
+
+def _generate_subnets_cidr_nat_rules():
+    env_list = []
+    for subnet in CONF.subnets:
+        s = CONF.get(subnet)
+        data_format = '"140 {direction} {name} cidr nat": ' \
+                      '{{"chain": "FORWARD", "{direction}": "{cidr}", ' \
+                      '"proto": "all", "action": "accept"}}'
+        env_list.append(data_format.format(
+            name=subnet, direction='destination', cidr=s.cidr))
+        env_list.append(data_format.format(
+            name=subnet, direction='source', cidr=s.cidr))
+    # Whitespace after newline required for indentation in templated yaml
+    return '\n  '.join(env_list)
+
+
 def _generate_environment(instack_root):
     """Generate an environment dict for instack
 
@@ -1223,10 +1389,16 @@ def _generate_environment(instack_root):
     else:
         raise RuntimeError('%s is not supported' % distro)
 
+    if CONF['additional_architectures']:
+        for arch in CONF['additional_architectures']:
+            env_name = ('enable_architecture_%s' % arch).upper()
+            instack_env[env_name] = six.text_type(True)
+
     # Convert conf opts to env values
     for opt in _opts:
         env_name = opt.name.upper()
         instack_env[env_name] = six.text_type(CONF[opt.name])
+
     # Opts that needs extra processing
     if CONF.inspection_runbench and not CONF.inspection_extras:
         raise RuntimeError('inspection_extras must be enabled for '
@@ -1249,6 +1421,10 @@ def _generate_environment(instack_root):
     instack_env['INSPECTION_KERNEL_ARGS'] = ' '.join(inspection_kernel_args)
 
     _process_drivers_and_hardware_types(instack_env)
+    instack_env['INSPECTION_SUBNETS'] = _generate_inspection_subnets()
+    instack_env['SUBNETS_CIDR_NAT_RULES'] = _generate_subnets_cidr_nat_rules()
+    instack_env['MASQUERADE_NETWORKS'] = _generate_masquerade_networks()
+    instack_env['SUBNETS_STATIC_ROUTES'] = _generate_subnets_static_routes()
 
     instack_env['SYSCTL_SETTINGS'] = _generate_sysctl_settings()
 
@@ -1596,7 +1772,7 @@ def _create_mistral_config_environment(instack_env, mistral):
     env_name = 'tripleo.undercloud-config'
     try:
         env_data = mistral.environments.get(env_name).variables
-    except ks_exceptions.NotFound:
+    except (ks_exceptions.NotFound, mistralclient_exc.APIException):
         # If the environment is not created, we need to
         # create it with the information in config_data
         mistral.environments.create(
@@ -1808,6 +1984,16 @@ def _post_config(instack_env, upgrade):
     ironic = ir_client.get_client(1, session=sess,
                                   os_ironic_api_version='1.21')
 
+    sdk = os_client_config.make_sdk(auth_url=auth_url,
+                                    project_name=project,
+                                    username=user,
+                                    password=password,
+                                    project_domain_name='Default',
+                                    user_domain_name='Default')
+
+    network = _ensure_neutron_network(sdk)
+    _config_neutron_segments_and_subnets(sdk, network.id)
+
     _configure_ssh_keys(nova)
     _ensure_ssh_selinux_permission()
     _delete_default_flavors(nova)
@@ -1848,6 +2034,213 @@ def _post_config(instack_env, upgrade):
         _migrate_to_convergence(heat)
 
 
+def _ensure_neutron_network(sdk):
+    try:
+        network = list(sdk.network.networks(name=PHYSICAL_NETWORK))
+        if not network:
+            network = sdk.network.create_network(
+                name=PHYSICAL_NETWORK, provider_network_type='flat',
+                provider_physical_network=PHYSICAL_NETWORK)
+            LOG.info("Network created %s", network)
+            # (hjensas) Delete the default segment, we create a new segment
+            # per subnet later.
+            segments = list(sdk.network.segments(network_id=network.id))
+            sdk.network.delete_segment(segments[0].id)
+            LOG.info("Default segment on network %s deleted.", network.name)
+        else:
+            LOG.info("Not creating %s network, because it already exists.",
+                     PHYSICAL_NETWORK)
+            network = network[0]
+    except Exception as e:
+        LOG.info("Network create/update failed %s", e)
+        raise
+
+    return network
+
+
+def _neutron_subnet_create(sdk, network_id, cidr, gateway, host_routes,
+                           allocation_pool, name, segment_id):
+    try:
+        # DHCP_START contains a ":" then assume a IPv6 subnet
+        if ':' in allocation_pool[0]['start']:
+            host_routes = ''
+            subnet = sdk.network.create_subnet(
+                name=name,
+                cidr=cidr,
+                gateway_ip=gateway,
+                host_routes=host_routes,
+                enable_dhcp=True,
+                ip_version='6',
+                ipv6_address_mode='dhcpv6-stateless',
+                ipv6_ra_mode='dhcpv6-stateless',
+                allocation_pools=allocation_pool,
+                network_id=network_id,
+                segment_id=segment_id)
+        else:
+            subnet = sdk.network.create_subnet(
+                name=name,
+                cidr=cidr,
+                gateway_ip=gateway,
+                host_routes=host_routes,
+                enable_dhcp=True,
+                ip_version='4',
+                allocation_pools=allocation_pool,
+                network_id=network_id,
+                segment_id=segment_id)
+        LOG.info("Subnet created %s", subnet)
+    except Exception as e:
+        LOG.error("Create subnet %s failed: %s", name, e)
+        raise
+
+    return subnet
+
+
+def _neutron_subnet_update(sdk, subnet_id, gateway, host_routes,
+                           allocation_pool, name):
+    try:
+        # DHCP_START contains a ":" then assume a IPv6 subnet
+        if ':' in allocation_pool[0]['start']:
+            host_routes = ''
+        subnet = sdk.network.update_subnet(
+                   subnet_id,
+                   name=name,
+                   gateway_ip=gateway,
+                   host_routes=host_routes,
+                   allocation_pools=allocation_pool)
+        LOG.info("Subnet updated %s", subnet)
+    except Exception as e:
+        LOG.error("Update subnet %s failed: %s", name, e)
+        raise
+
+
+def _neutron_segment_create(sdk, name, network_id, phynet):
+    try:
+        segment = sdk.network.create_segment(
+                    name=name,
+                    network_id=network_id,
+                    physical_network=phynet,
+                    network_type='flat')
+        LOG.info("Neutron Segment created %s", segment)
+    except Exception as e:
+        LOG.info("Neutron Segment %s create failed %s", name, e)
+        raise
+
+    return segment
+
+
+def _neutron_segment_update(sdk, segment_id, name):
+    try:
+        segment = sdk.network.update_segment(segment_id, name=name)
+        LOG.info("Neutron Segment updated %s", segment)
+    except Exception as e:
+        LOG.info("Neutron Segment %s update failed %s", name, e)
+        raise
+
+
+def _ensure_neutron_router(sdk, name, subnet_id):
+    try:
+        router = sdk.network.create_router(name=name, admin_state_up='true')
+        sdk.network.add_interface_to_router(router.id, subnet_id=subnet_id)
+    except Exception as e:
+        LOG.error("Create router for subnet %s failed: %s", name, e)
+        raise
+
+
+def _get_subnet(sdk, cidr, network_id):
+    try:
+        subnet = list(sdk.network.subnets(cidr=cidr, network_id=network_id))
+    except Exception:
+        raise
+
+    return False if not subnet else subnet[0]
+
+
+def _get_segment(sdk, phy, network_id):
+    try:
+        segment = list(sdk.network.segments(physical_network=phy,
+                                            network_id=network_id))
+    except Exception:
+        raise
+
+    return False if not segment else segment[0]
+
+
+def _config_neutron_segments_and_subnets(sdk, ctlplane_id):
+    s = CONF.get(CONF.local_subnet)
+    subnet = _get_subnet(sdk, s.cidr, ctlplane_id)
+    if subnet and not subnet.segment_id:
+        LOG.warn("Local subnet %s already exists and is not associated with a "
+                 "network segment. Any additional subnets will be ignored.",
+                 CONF.local_subnet)
+        host_routes = [{'destination': '169.254.169.254/32',
+                        'nexthop': str(netaddr.IPNetwork(CONF.local_ip).ip)}]
+        allocation_pool = [{'start': s.dhcp_start, 'end': s.dhcp_end}]
+        _neutron_subnet_update(sdk, subnet.id, s.gateway, host_routes,
+                               allocation_pool, CONF.local_subnet)
+        # If the subnet is IPv6 we need to start a router so that router
+        # advertisments are sent out for stateless IP addressing to work.
+        if ':' in s.dhcp_start:
+            _ensure_neutron_router(sdk, CONF.local_subnet, subnet.id)
+    else:
+        for name in CONF.subnets:
+            s = CONF.get(name)
+
+            phynet = name
+            if name == CONF.local_subnet:
+                phynet = PHYSICAL_NETWORK
+
+            metadata_nexthop = s.gateway
+            if str(netaddr.IPNetwork(CONF.local_ip).ip) in s.cidr:
+                metadata_nexthop = str(netaddr.IPNetwork(CONF.local_ip).ip)
+
+            host_routes = [{'destination': '169.254.169.254/32',
+                            'nexthop': metadata_nexthop}]
+            allocation_pool = [{'start': s.dhcp_start, 'end': s.dhcp_end}]
+
+            subnet = _get_subnet(sdk, s.cidr, ctlplane_id)
+            segment = _get_segment(sdk, phynet, ctlplane_id)
+
+            if name == CONF.local_subnet:
+                if ((subnet and not segment) or
+                   (subnet and segment and subnet.segment_id != segment.id)):
+                    LOG.error(
+                      'The cidr: %s of the local subnet is already used in '
+                      'subnet: %s associated with segment_id: %s.' %
+                      (s.cidr, subnet.id, subnet.segment_id))
+                    raise RuntimeError('Local subnet cidr already associated.')
+
+            if subnet:
+                _neutron_segment_update(sdk, subnet.segment_id, name)
+                _neutron_subnet_update(sdk, subnet.id, s.gateway, host_routes,
+                                       allocation_pool, name)
+            else:
+                if segment:
+                    _neutron_segment_update(sdk, segment.id, name)
+                else:
+                    segment = _neutron_segment_create(sdk, name,
+                                                      ctlplane_id, phynet)
+                if CONF.enable_routed_networks:
+                    subnet = _neutron_subnet_create(sdk, ctlplane_id, s.cidr,
+                                                    s.gateway, host_routes,
+                                                    allocation_pool, name,
+                                                    segment.id)
+                elif name == CONF.local_subnet:
+                    # Create subnet with segment_id: None if routed networks
+                    # is not enabled.
+                    # TODO(hjensas): Deprecate option and remove this once
+                    # tripleo-ui can support managing baremetal port
+                    # attributes.
+                    subnet = _neutron_subnet_create(sdk, ctlplane_id, s.cidr,
+                                                    s.gateway, host_routes,
+                                                    allocation_pool, name,
+                                                    None)
+
+            # If the subnet is IPv6 we need to start a router so that router
+            # advertisments are sent out for stateless IP addressing to work.
+            if ':' in s.dhcp_start:
+                _ensure_neutron_router(sdk, name, subnet.id)
+
+
 def _handle_upgrade_fact(upgrade=False):
     """Create an upgrade fact for use in puppet
 
@@ -1900,6 +2293,10 @@ def install(instack_root, upgrade=False):
         _configure_logging(DEFAULT_LOG_LEVEL, PATHS.LOG_FILE)
         LOG.info('Logging to %s', PATHS.LOG_FILE)
         _load_config()
+        _load_subnets_config_groups()
+        # Since 'subnets' parameter in opts is used to dynamically
+        # add per network groups, re-load config.
+        _load_config()
         _clean_os_refresh_config()
         _clean_os_collect_config()
         _validate_configuration()
